#include "config.h"
#include <limits.h>
#include "UIRsrcMgr.h"
#include "VarNavGuide.h"
#include "InfoBar.h"
#include "Item.h"
#include "DlgBtn.h"
#include "vncviewer.h"
#include "debug.h"
#include "SDL_stretch.h"
#include "Dialog.h"

const int BTN_WIDTH     = 188;
const int BTN_HEIGHT    = 34;
const int BTN_SPACING   = 50;
const int BTN_THIN      = 3;
const int BTN_TOP_SPACE = 18;


Dialog::
Dialog		(SDL_Surface		*screen,
			 const ::SDL_Rect&	rect)
    : BaseDialog(screen,rect)
{
	m_AlphaValue	= (UCHAR_MAX * 9) / 10; // UCHAR_MAX * 0.9
	m_NavGuide		= NULL;
	m_InfoBar		= NULL;
}

Dialog::
~Dialog()
{
	delete m_NavGuide;
	delete m_InfoBar;
	RestoreOrigImage();
	m_restore = false;
    SDL_UpdateRect(m_screen, 0,0,0,0);
}


int Dialog::
Init	()
{    
    int result;
	result = BaseDialog::Init();
	if (result != 0) {
		return result;
	}

    Grain *p;
    SDL_Rect rect;

    // Create the Exit button (left of center dialog)
    rect.w = BTN_WIDTH + BTN_THIN * 2;
    rect.h = BTN_HEIGHT + BTN_THIN * 2;
    rect.x = (drawable_area.w/2) - (BTN_SPACING/2) - rect.w;
    rect.y = drawable_area.y + BTN_TOP_SPACE;
    p = new DlgBtn(m_screen, rect, DlgBtn::BtnExit);
    AddGrain(p);
    result = p->Init();
    if (result != 0) {
        return result;
    }

    // Create the Continue button (right of center dialog)
    rect.x = drawable_area.w / 2 + BTN_SPACING / 2;
    p = new DlgBtn(m_screen, rect, DlgBtn::BtnContinue);
    AddGrain(p);
    result = p->Init();
    if (result != 0) {
        return result;
    }

    // Select the first (i.e. exit) button
	assert(dynamic_cast<DlgBtn *>(m_child));
	static_cast<DlgBtn *>(m_child)->Select();

	// setup infobar
    rect = INFOBAR_DEFAULT_RECT;
    rect.w = m_screen->w;
	rect.y = drawable_area.y - rect.h;
	m_InfoBar = new InfoBar(m_screen, rect); 
	if (!m_InfoBar) {
		return -1;
	}	
	result = m_InfoBar->Init();
	if (result != 0) {
		return result;
	}	

	// setup NavGuide
    rect = NAVGUIDE_DEFAULT_RECT;
    rect.w = m_screen->w;
    rect.y = m_screen->h - 42;
	m_NavGuide = new VarNavGuide(m_screen, rect);
	if (!m_NavGuide) {
		return -1;
	}
	result = m_NavGuide->Init();
	if (result != 0) {
		return result;
	}
	result = SetNavGuide();
	if (result != 0) {
		return result;
	}

    UpdateResources();

	return 0;
}

int Dialog::
SetNavGuide	()
{
	assert(m_NavGuide);

    int result;
    NavGuide *p = m_NavGuide;
	result = p->SetIconDescr(NavGuide::Icon1, NavItem::IconEnter, NavItem::DescrDecide);
	if (result != 0) { 
		return result;
	}
	result = p->SetIconDescr(NavGuide::Icon2, NavItem::IconHoriz, NavItem::DescrSelect);
	if (result != 0) { 
		return result;
	}
	result = p->SetIconDescr(NavGuide::Icon3, NavItem::IconESC,NavItem::DescrBack);
	if (result != 0) { 
		return result;
	}
	result = p->SetIconDescr(NavGuide::Icon4, NavItem::IconHidden,NavItem::DescrDecide);
	if (result != 0) { 
		return result;
	}
	result = p->SetIconDescr(NavGuide::Icon5, NavItem::IconHidden,NavItem::DescrDecide);
	if (result != 0) { 
		return result;
	}
	return 0;
}

int Dialog::
Draw	()
{
	assert(m_NavGuide);
	assert(m_InfoBar);

    int result;

	// Get a resrouce from UIRsrcMgr
    SDL_Surface *surface;
	surface = UIRsrcMgr::Instance()->GetResource(UIRsrcMgr::dlgBackground);
	if (!surface) {
		return -1;
	}

	// Set alpha value
	result = SDL_SetAlpha(surface, m_AlphaValue != UCHAR_MAX ? SDL_SRCALPHA : 0, m_AlphaValue);
	if (result != 0) {
		SDL_ERROR();
		return result;
	}

	// Blit the resource to m_screen
	result = SDL_StretchBlit(surface,   NULL, 
                             m_screen,  &drawable_area,
                             SDL_STRETCH_BILINEAR);
	if (result != 0) {
		SDL_ERROR();
		return result;
	}

	// draw each item
    Grain *p = m_child;
	while (p) {
		result = p->Draw();
		if (result != 0) {
			return result;
		}
		p = p->GetNext();
	}

	result = m_InfoBar->Draw();
	if (result != 0) {
		return result;
	}

	result = m_NavGuide->Draw();
	if (result != 0) {
		return result;
	}
	

	return 0;
}


void Dialog::
Update		()
{
	assert(m_NavGuide);
	assert(m_InfoBar);

	m_NavGuide->Update();
	m_InfoBar->Update();
	UpdateRect(m_screen, drawable_area);
}

ProcStatus Dialog::
Event	(const ::SDL_Event& event)
{
    Grain *p = m_child;
	while (p) {
		assert(dynamic_cast<DlgBtn *>(p));
		if (static_cast<DlgBtn*>(p)->IsSelected()) {
            DebugPrint("%s: Passing event to child widget\n", __FUNCTION__);
			return p->Event(event);
		}
		p = p->GetNext();
	}
	return ProcKeepGoing;
}


void Dialog::
DoCallback  (const DlgBtn::BtnType type)
{
    DebugPrint("Entering %s\n", __FUNCTION__);
    if ( type == DlgBtn::BtnExit ) {
        DebugPrint("Exit button selected.  Closing VNC\n");
        ::Cleanup(CLIENT_QUIT);
    } else if ( type == DlgBtn::BtnContinue ) {
        DebugPrint("Continue button selected.  Closing dialog window\n");
        ::CloseDialog();
    }
}

void
Dialog::UpdateResources()
{
    SDL_Surface* infoImage = 
        UIRsrcMgr::Instance()->GetResource(UIRsrcMgr::msgBarBack1);

    if ( infoImage ) {
        m_InfoBar->SetBackground(infoImage);
    }
    infoImage = UIRsrcMgr::Instance()->GetResource(UIRsrcMgr::labGuide3);
    if ( infoImage ) {
        SDL_Rect rect;
        rect.w = infoImage->w; rect.h = infoImage->h;        
        rect.x = drawable_area.w/2 - rect.w/2; rect.y = 33 - rect.h/2;
        m_InfoBar->SetForeground(infoImage, rect);
    }
}
