/*
 *  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

/*
 * argsresources.c - deal with command-line args and resources.
 */

#include "vncviewer.h"


/*
 * vncServerHost and vncServerPort are set either from the command line or
 * from a dialog box.
 */

char vncServerHost[256];
int vncServerPort = 0;


/*
 * appData is our application-specific data which can be set by the user with
 * application resource specs.  The AppData structure is defined in the header
 * file.
 */

AppData appData;


/*
 * removeArgs() is used to remove some of command line arguments.
 */

void
removeArgs(int *argc, char** argv, int idx, int nargs)
{
    int i;
    if ((idx+nargs) > *argc) return;

#ifndef NDEBUG
    fprintf(stderr, "Removing commandline: %s\n", argv[idx]);
    for (i = idx+1; i < (idx+nargs); i++) {
        //fprintf(stderr, "\t parameter: %s\n", argv[i]);
    }
#endif

    for (i = idx+nargs; i < *argc; i++) {
        argv[i-nargs] = argv[i];
    }
    *argc -= nargs;
}


/*
 * usage() prints out the usage message.
 */
void
usage(void)
{
    fprintf(stderr,
            "TightVNC viewer version 1.2.2 (based on VNC 3.3.3r2)\n"
            "\n"
            "Usage: %s [<OPTIONS>] [<HOST>][:<DISPLAY#>]\n"
            "       %s [<OPTIONS>] -listen [<DISPLAY#>]\n"
            "\n"
            "<OPTIONS> are:\n"
            "        -fullscreen\n"
            "        -encodings <ENCODING-LIST> (e.g. \"tight copyrect\")\n"
            "        -depth <DEPTH>\n"
            "        -compresslevel <COMPRESS-VALUE> (0..9: 0-fast, 9-best)\n"
            "        -quality <JPEG-QUALITY-VALUE> (0..9: 0-low, 9-high)\n"
            "        -encrypt <PASSWORD>\n"
            "        -udp\n"
            "        -width <SERVER-WIDTH>\n"
            "        -height <SERVER-HEIGHT>\n"
            "        -server  <SERVER-IP>\n"
            "\n"
            "Option names may be abbreviated, e.g. -bgr instead of -bgr233.\n"
            "\n", programName, programName);
    Cleanup(1);
}


void
SetAppDataDefaults()
{
    appData.shareDesktop = SDL_TRUE;
    appData.viewOnly = SDL_TRUE;
    appData.fullScreen = SDL_FALSE;
    appData.raiseOnBeep = SDL_TRUE;

    appData.encodingsString = 0;
    appData.encryptString = 0;

    appData.useBGR233 = SDL_FALSE;
    appData.nColours = 256;
    appData.useSharedColours = SDL_TRUE;
    appData.forceOwnCmap = SDL_FALSE;
    appData.forceTrueColour = SDL_FALSE;
    appData.requestedDepth = 0;

    appData.useShm = SDL_TRUE;

    appData.wmDecorationWidth = 4;
    appData.wmDecorationHeight = 24;

    appData.passwordFile = 0;
    appData.passwordDialog = SDL_FALSE;

    appData.debug = SDL_FALSE;

    appData.rawDelay = 0;
    appData.copyRectDelay = 0;

    appData.bumpScrollTime = 25;
    appData.bumpScrollPixels = 20;

    appData.compressLevel = -1;
    appData.qualityLevel = -1;
    appData.useRemoteCursor = SDL_TRUE;

    appData.useUdpSocket = SDL_FALSE;

    // default 1024x768x32 bpp
    si.framebufferWidth   = 1024;
    si.framebufferHeight  = 768;
    si.format.bitsPerPixel= 32;
    si.format.depth       = 32;
    si.format.redShift    = 16;
    si.format.greenShift  = 8;
    si.format.blueShift   = 0;
    si.format.redMax      = 255;
    si.format.greenMax    = 255;
    si.format.blueMax     = 255;
    si.format.trueColour  = SDL_TRUE;
}

/*
 * GetArgsAndResources() deals with resources and any command-line arguments
 * not already processed by XtVaAppInitialize().  It sets vncServerHost and
 * vncServerPort and all the fields in appData.
 */

void
GetArgsAndResources(int argc, char **argv)
{
    int i;
    char *vncServerName = NULL;

#ifdef BENCHMARK
    decryptTime = 0;
    decryptBytes = 0;
    encryptTime = 0;
    encryptBytes = 0;
#endif

    /* Initialize the default app data parameters */
    SetAppDataDefaults();

    /* Parse the command line parameters in to the appdata structure*/
    i = 1;
    while(i < argc) {
        if (strcasecmp("-compresslevel", argv[i]) == 0) {
            if ( (i+1) < argc) {
                appData.compressLevel = atoi(argv[i+1]);
                removeArgs(&argc, argv, i, 2);
            } else {
                usage();
            }

        } else if (strcasecmp("-quality", argv[i]) == 0) {
            if ( (i+1) < argc) {
                appData.qualityLevel = atoi(argv[i+1]);
                removeArgs(&argc, argv, i, 2);
            } else {
                usage();
            }

        } else if (strcasecmp("-encodings", argv[i]) == 0) {
            if ( (i+1) < argc) {
                appData.encodingsString = strdup(argv[i+1]);
                removeArgs(&argc, argv, i, 2);
            } else {
                usage();
            }

        } else if (strcasecmp("-width", argv[i]) == 0) {
            if ( (i+1) < argc) {
                si.framebufferWidth = atoi(argv[i+1]);
                removeArgs(&argc, argv, i, 2);
            } else {
                usage();
            }

        } else if (strcasecmp("-height", argv[i]) == 0) {
            if ( (i+1) < argc) {
                si.framebufferHeight = atoi(argv[i+1]);
                removeArgs(&argc, argv, i, 2);
            } else {
                usage();
            }

        } else if (strcasecmp("-depth", argv[i]) == 0) {
            if ( (i+1) < argc) {
                visdepth = atoi(argv[i+1]); 
                visbpp = visdepth;
                si.format.bitsPerPixel= visdepth;
                si.format.depth       = visdepth;
                removeArgs(&argc, argv, i, 2);
            } else {
                usage();
            }

        } else if (strcasecmp(argv[i], "-encrypt") == 0) {
            if ( (i+1) < argc ) {
                appData.encryptString = argv[i+1];
                g_UseEncryption = SDL_TRUE;
                removeArgs(&argc, argv, i, 2);
            } else {
                usage();
            }

        } else if (strcasecmp(argv[i], "-server") == 0) {
            if ( (i+1) < argc ) {
                int a;
                int addr[4];
                sscanf(argv[i+1], "%u.%u.%u.%u", 
                       &addr[0], &addr[1], &addr[2], &addr[3]);

                for ( a = 0; a < 4; a++ ) {
                    appData.serverIP[a] = (char)(addr[a] & 0xFF);
                }

                removeArgs(&argc, argv, i, 2);
            } else {
                usage();
            }

        } else if (strcasecmp(argv[i], "-shifts") == 0) {
            if ( (i+1) < argc ) {
                Uint32 red, green, blue;
                sscanf(argv[i+1], "%u,%u,%u", &red, &green, &blue);

                si.format.redShift = (Uint8)red;
                si.format.greenShift = (Uint8)green;
                si.format.blueShift = (Uint8)blue;

                removeArgs(&argc, argv, i, 2);
            } else {
                usage();
            }

        } else if (strcasecmp(argv[i], "-max") == 0) {
            if ( (i+1) < argc ) {
                Uint32 red, green, blue;
                sscanf(argv[i+1], "%u,%u,%u", &red, &green, &blue);

                si.format.redMax = red;
                si.format.greenMax = green;
                si.format.blueMax = blue;

                removeArgs(&argc, argv, i, 2);
            } else {
                usage();
            }

        } else if (strcasecmp("-fullscreen", argv[i]) == 0) {
            appData.fullScreen = SDL_TRUE;
            removeArgs(&argc, argv, i, 1);

        } else if (strcasecmp("-udp", argv[i]) == 0) {
            if ( (i+1) < argc) {
                if ( atoi(argv[i+1]) ) {
                    appData.useUdpSocket = SDL_TRUE;
                    g_UseUdpSocket = SDL_TRUE; 
                } else {
                    appData.useUdpSocket = SDL_FALSE;
                    g_UseUdpSocket = SDL_FALSE; 
                }
                removeArgs(&argc, argv, i, 2);
            } else {
                usage();
            }

        } else {
            fprintf(stderr, "Bad VNC commandline option: %s\n",argv[i]);
            usage();
        }
    }

#ifdef FORCE_UDP
    fprintf(stderr, "*** Forcing UDP mode\n");
    appData.useUdpSocket = SDL_TRUE;
    g_UseUdpSocket = SDL_TRUE; 
#endif

    /* Check any remaining command-line arguments.  If -listen was specified
       there should be none.  Otherwise the only argument should be the VNC
       server name.  If not given then pop up a dialog box and wait for the
       server name to be entered. */

    if (listenSpecified) {
        if (argc != 1) {
            fprintf(stderr,"\n%s -listen: invalid command line argument: %s\n",
                    programName, argv[1]);
            usage();
        }
        return;
    }

    if (argc == 1) {
        fprintf(stderr,"\n%s -listen: invalid command line argument: %s\n",
                programName, argv[1]);
        usage();
        return;
        
    } else if (argc != 2) {
        usage();
    } else {
        vncServerName = argv[1];

        if (vncServerName[0] == '-')
            usage();
    }

    if (strlen(vncServerName) > 255) {
        fprintf(stderr,"VNC server name too long\n");
        Cleanup(1);
    }

    for (i = 0; vncServerName[i] != ':' && vncServerName[i] != 0; i++);

    strncpy(vncServerHost, vncServerName, i);

    if (vncServerName[i] == ':') {
        vncServerPort = atoi(&vncServerName[i+1]);
    } else {
        vncServerPort = 0;
    }

    if (vncServerPort < 100)
        vncServerPort += SERVER_PORT_OFFSET;

    if ( appData.encryptString ) {
        printf("Using encryption string: %8.8s\n", appData.encryptString);
        g_UseEncryption = SDL_TRUE;
    } else {
        printf("Not using encryption\n");
        g_UseEncryption = SDL_FALSE;
    }
}
