/*
 *  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

/*
 * desktop.c - functions to deal with "desktop" window.
 */

#include <vncviewer.h>
#include <SDL.h>

static SDL_Cursor* CreateDotCursor();
static void CopyBGR233ToScreen(Uint8 *buf, int x, int y, int width,int height);

#ifndef NDEBUG
static void PrintVideoInfo();
#endif

/*
 * DesktopInitBeforeRealization creates the "desktop" widget and the viewport
 * which controls it.
 */

void
DesktopInitBeforeRealization()
{
    /* Don't show the local cursor */
    SDL_ShowCursor(0);

    /* Finally try to get the input focus.  With some WMs we might
       have to grab the keyboard, but this seems to be OK with the
       ones I've tried. */
    SDL_WM_GrabInput(SDL_GRAB_ON);    

    /* TODO: Find the closest resolution */
    if (!dpy) {
        SDL_Surface* image;
        int flags;

#ifndef NDEBUG
        fprintf(stderr, "Best video mode:\n");
        PrintVideoInfo();
#endif

        flags = SDL_FULLSCREEN|SDL_SWSURFACE;
        image = SDL_SetVideoMode(si.framebufferWidth, 
                                 si.framebufferHeight,
                                 visdepth, flags);

        if ( image == NULL ) {
            Cleanup(-1);
        }

#ifndef NDEBUG
        fprintf(stderr, "Using video mode:\n");
        PrintVideoInfo();
        fprintf(stderr, "\tSurface BPP: %d\n", image->format->BitsPerPixel);
        fprintf(stderr, "\tSurface Dimensions: %d x %d\n", image->w, image->h);
        fprintf(stderr, "\tSurface pitch: %d bytes\n", image->pitch);
#endif

        /* Set the colormap */
        if ( image->format->palette ) {
            SDL_SetPalette(image, SDL_LOGPAL|SDL_PHYSPAL, cmap.colors, 
                           0, cmap.ncolors);
        }

        dpy = image;        
    }
}


/*
 * DesktopInitAfterRealization does things which require the X windows to
 * exist.  It creates some GCs and sets the dot cursor.
 */

void
DesktopInitAfterRealization()
{
    Uint32 colour;
    SDL_Cursor* dotCursor;

    /* Fill the background with black */
    colour = SDL_MapRGB(dpy->format, 0, 0, 0);
    SDL_FillRect(dpy, NULL, colour);

    /* Set the cursor */
    dotCursor = CreateDotCursor();
    SDL_SetCursor(dotCursor);

    SDL_Flip(dpy);
}


/*
 * HandleBasicDesktopEvent - deal with expose and leave events.
 */

void
HandleBasicDesktopEvent(SDL_Event* event)
{
    switch (event->type) {

    case SDL_VIDEOEXPOSE:
        SendFramebufferUpdateRequest(0, 0, dpy->w, dpy->h, SDL_FALSE);
        break;

    case SDL_ACTIVEEVENT:
        {
            if ( event->active.gain ) {
                SendFramebufferUpdateRequest(0, 0, dpy->w, dpy->h, SDL_FALSE);
            }
        }
        break;
    }
}


/*
 * SendRFBEvent is an action which sends an RFB event.  It can be used in two
 * ways.  Without any parameters it simply sends an RFB event corresponding to
 * the X event which caused it to be called.  With parameters, it generates a
 * "fake" RFB event based on those parameters.  The first parameter is the
 * event type, either "ptr", "keydown", "keyup" or "key" (down&up).  For a
 * "key" event the second parameter is simply a keysym string as understood by
 * XStringToKeysym().  For a "ptr" event, the following three parameters are
 * just X, Y and the button mask (0 for all up, 1 for button1 down, 2 for
 * button2 down, 3 for both, etc).
 */

void
SendRFBEvent(SDL_Event* ev, String *params, int *num_params)
{
    if (appData.fullScreen && ev->type == SDL_MOUSEMOTION) {
        if (BumpScroll(ev))
            return;
    }

#if 0
    if ( ev->type == SDL_KEYDOWN && ev->key.keysym.sym == SDLK_ESCAPE ) {
        OpenDialog();
        return;
    }

    if (appData.viewOnly) return;
#endif

    switch (ev->type) {
        Uint8 buttons;
        Uint32 keys;

    case SDL_MOUSEMOTION:
        while (SDL_PeepEvents(ev, 1, SDL_GETEVENT, SDL_MOUSEMOTIONMASK) == 1)
            ;	/* discard all queued motion notify events */

        SendPointerEvent(ev->motion.x, ev->motion.y, ev->motion.state);
        break;

    case SDL_MOUSEBUTTONDOWN:
        buttons = ev->button.state << ev->button.button;
        SendPointerEvent(ev->button.x, ev->button.y, buttons);
        break;

    case SDL_MOUSEBUTTONUP:
        buttons = ev->button.state << ev->button.button;
        SendPointerEvent(ev->button.x, ev->button.y, buttons);
        break;

    case SDL_KEYDOWN:
    case SDL_KEYUP:
        keys = ev->key.keysym.sym;
        SendKeyEvent(keys, (ev->type == SDL_KEYDOWN));
        break;

    default:
#ifndef NDEBUG
        fprintf(stderr,"Invalid event passed to SendRFBEvent\n");
#endif
        break;
    }
}


/*
 * CreateDotCursor.
 */

static SDL_Cursor*
CreateDotCursor()
{
    SDL_Cursor* cursor;
    static char srcBits[] = { 0, 14,14,14, 0 };
    static char mskBits[] = { 14,31,31,31,14 };

    cursor = SDL_CreateCursor(srcBits, mskBits, 8, 5, 5, 2);
    return cursor;
}


/*
 * CopyDataToScreen.
 */

void
CopyDataToScreen(char *buf, int x, int y, int width, int height)
{
    if (appData.rawDelay != 0) {
        SDL_Rect rect = { x, y, width, height };
        SDL_FillRect(dpy, &rect, SDL_MapRGB(dpy->format, 0,0,0));
        SDL_UpdateRect(dpy, x, y, width, height);
        usleep(appData.rawDelay * 1000);
    }

    if (!appData.useBGR233) {
        int h;
        int widthInBytes = width * (si.format.bitsPerPixel / 8);
        int scrnWidthInBytes = dpy->pitch;
        char* scrn;

#ifdef ARGB1555
            int w;
            Uint16* src;
            Uint16* dst;
#endif

        if (SDL_MUSTLOCK(dpy)) {
            SDL_LockSurface(dpy);
        }

        scrn = ((Uint8*)dpy->pixels + 
               y * scrnWidthInBytes + 
               x * dpy->format->BytesPerPixel);

        for (h = 0; h < height; h++) {
#ifdef ARGB1555
            src = (Uint16*)buf;
            dst = (Uint16*)scrn;
            for( w = 0; w < width; w++ ) {
                dst[w] = (src[w] & 0x001F) | ((src[w] >> 1) & 0x7FE0);
            }
#else
            memcpy(scrn, buf, widthInBytes);
#endif

            buf += widthInBytes;
            scrn += scrnWidthInBytes;
        }

        if (SDL_MUSTLOCK(dpy)) {
            SDL_UnlockSurface(dpy);
        }

    } else {
        CopyBGR233ToScreen((Uint8 *)buf, x, y, width, height);
    }

#ifdef MITSHM
    if (appData.useShm) {
        XShmPutDpy(dpy, desktopWin, gc, dpy, x, y, x, y, width, height, SDL_FALSE);
        return;
    }
#endif
}


/*
 * CopyBGR233ToScreen.
 */

static void
CopyBGR233ToScreen(Uint8 *buf, int x, int y, int width, int height)
{
    int p, q;
    int xoff = 7 - (x & 7);
    int xcur;
    int fbwb = si.framebufferWidth / 8;
    Uint8 *scr1 = ((Uint8 *)dpy->pixels) + y * fbwb + x / 8;
    Uint8 *scrt;
    Uint8 *scr8 = ((Uint8 *)dpy->pixels) + y * si.framebufferWidth + x;
    Uint16 *scr16 = ((Uint16 *)dpy->pixels) + y * si.framebufferWidth + x;
    Uint32 *scr32 = ((Uint32 *)dpy->pixels) + y * si.framebufferWidth + x;

    switch (visbpp) {

        /* thanks to Chris Hooper for single bpp support */

    case 1:
        for (q = 0; q < height; q++) {
            xcur = xoff;
            scrt = scr1;
            for (p = 0; p < width; p++) {
                *scrt = ((*scrt & ~(1 << xcur))
                         | (BGR233ToPixel[*(buf++)] << xcur));

                if (xcur-- == 0) {
                    xcur = 7;
                    scrt++;
                }
            }
            scr1 += fbwb;
        }
        break;

    case 8:
        for (q = 0; q < height; q++) {
            for (p = 0; p < width; p++) {
                *(scr8++) = BGR233ToPixel[*(buf++)];
            }
            scr8 += si.framebufferWidth - width;
        }
        break;

    case 16:
        for (q = 0; q < height; q++) {
            for (p = 0; p < width; p++) {
                *(scr16++) = BGR233ToPixel[*(buf++)];
            }
            scr16 += si.framebufferWidth - width;
        }
        break;

    case 32:
        for (q = 0; q < height; q++) {
            for (p = 0; p < width; p++) {
                *(scr32++) = BGR233ToPixel[*(buf++)];
            }
            scr32 += si.framebufferWidth - width;
        }
        break;
    }
}

#ifndef NDEBUG
static void
PrintVideoInfo()
{
    const SDL_VideoInfo* vinfo;
    char  driver[256];

    vinfo = SDL_GetVideoInfo();
    SDL_VideoDriverName(driver, 256);

    fprintf(stderr, "Video Info for driver: %s\n", driver);
    fprintf(stderr, "\tHW available: %d\n", vinfo->hw_available);
    fprintf(stderr, "\tHW blit: %d\n", vinfo->blit_hw);
    fprintf(stderr, "\tHW blit CC: %d\n", vinfo->blit_hw_CC);
    fprintf(stderr, "\tHW blit Alpha: %d\n", vinfo->blit_hw_A);
    fprintf(stderr, "\tSW blit: %d\n", vinfo->blit_sw);
    fprintf(stderr, "\tSW blit CC: %d\n", vinfo->blit_sw_CC);
    fprintf(stderr, "\tSW blit Alpha: %d\n", vinfo->blit_sw_A);
    fprintf(stderr, "\tBytes per pixel: %d\n", vinfo->vfmt->BytesPerPixel);
    fprintf(stderr, "\tBits per pixel: %d\n", vinfo->vfmt->BitsPerPixel);
}
#endif
