/*
 *  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

/*
 * listen.c - listen for incoming connections
 */

#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/utsname.h>
#include <vncviewer.h>
#include "vnes.h"

/* Timeout after 10 seconds have elapsed */
#define LISTEN_TIMEOUT      10

SDL_bool listenSpecified = SDL_FALSE;
int listenPort = 0;

/*
 * listenForIncomingConnections() - listen for incoming connections from
 * servers, and fork a new process to deal with each connection.  We must do
 * all this before invoking any Xt functions - this is because Xt doesn't
 * cope with forking very well.
 */
void
listenForIncomingConnections(int *argc, char **argv, int listenArgIndex)
{
    int cnt;
    int listenSocket;
    fd_set fds;

    listenSpecified = SDL_TRUE;

    if (listenArgIndex+1 < *argc && argv[listenArgIndex+1][0] >= '0' &&
        argv[listenArgIndex+1][0] <= '9') {

        listenPort = LISTEN_PORT_OFFSET + atoi(argv[listenArgIndex+1]);
        removeArgs(argc, argv, listenArgIndex, 2);

    } else {

        fprintf(stderr,"%s: cannot work out which display number to "
                "listen on.\n", programName);
        fprintf(stderr,"Please specify explicitly with -listen <num>\n");
        Cleanup(1);
    }

    // Search the command line for the udp option
    cnt = 1;
    while (cnt < *argc) {
        if (strcasecmp("-udp", argv[cnt]) == 0) {
            if ( (cnt+1) < *argc) {
                if ( atoi(argv[cnt+1]) ) {
                    fprintf(stderr, "VNC: Using UDP\n");
                    appData.useUdpSocket = SDL_TRUE;
                    g_UseUdpSocket = SDL_TRUE; 
                } else {
                    fprintf(stderr, "VNC: Using TCP\n");
                    appData.useUdpSocket = SDL_FALSE;
                    g_UseUdpSocket = SDL_FALSE; 
                }
                removeArgs(argc, argv, cnt, 2);

            } else {
                usage();
            }

        } else if (strcasecmp(argv[cnt], "-server") == 0) {
            if ( (cnt+1) < *argc ) {
                int a;
                int addr[4];
                sscanf(argv[cnt+1], "%u.%u.%u.%u", 
                       &addr[0], &addr[1], &addr[2], &addr[3]);

#ifndef NDEBUG
                fprintf(stderr, "*** Recieved server IP: ");
#endif
                for ( a = 0; a < 4; a++ ) {
                    appData.serverIP[a] = (char)(addr[a] & 0xFF);
#ifndef NDEBUG
                    fprintf(stderr, "%u.", addr[a]);
#endif
                }
#ifndef NDEBUG
                fprintf(stderr, "\n");
#endif

                removeArgs(argc, argv, cnt, 2);
            } else {
                usage();
            }

        } else {
            cnt++;
        }
    }


#ifdef FORCE_UDP
    fprintf(stderr, "*** Forcing UDP mode\n");
    appData.useUdpSocket = SDL_TRUE;
    g_UseUdpSocket = SDL_TRUE; 
#endif

    if ( appData.useUdpSocket ) {
        listenSocket = ListenAtUdpPort(listenPort);
        fprintf(stderr,"%s -listen: Listening on UDP port %d\n",
                programName,listenPort);
    } else {
        listenSocket = ListenAtTcpPort(listenPort);
        fprintf(stderr,"%s -listen: Listening on TCP port %d\n",
                programName,listenPort);
    }
    if (listenSocket < 0) Cleanup(1);

    while (SDL_TRUE) {
        struct timeval tm;
        int ret;
        SDL_Event event;
        int status, pid;
        vnes_msg_t vnesmsg;

        fprintf(stderr,"VNC: Waiting for connection...\n");       

        /* reap any zombies */
        while ((pid= wait3(&status, WNOHANG, (struct rusage *)0))>0);

        /* Send the ready message to the controlling daemon */
        memset(&vnesmsg, 0, sizeof(vnesmsg));
        vnesmsg.type = VNES_MSG_VNC_INITIALIZED;
        if (SendDomainMessage(&vnesmsg) == 0) {
            // SendDomainMessage failed so exit with error code 1
            Cleanup(1);            
        }

        tm.tv_sec = LISTEN_TIMEOUT;
        tm.tv_usec = 0;
        FD_ZERO(&fds); 
        FD_SET(listenSocket, &fds);
#ifdef NOUSE_LISTEN_TIMEOUT
        ret = select(FD_SETSIZE, &fds, NULL, NULL, NULL);
#else
        ret = select(FD_SETSIZE, &fds, NULL, NULL, &tm);
#endif
        if ( ret <= 0 ) {
            close(listenSocket);

            if ( ret == -1 ) {
                perror("VNC listen");

            }
            Cleanup(1);
        }

        // initialize the socket descriptors
        rfbsockTCP = -1;
        rfbsockUDP = -1;

        /* discard any events */
        while (SDL_PollEvent(&event)) {};

        if (FD_ISSET(listenSocket, &fds)) {
            if ( appData.useUdpSocket ) {
                rfbsockUDP = listenSocket;
                if (!SetNonBlocking(rfbsockUDP)) {
                    Cleanup(1);
                }

            } else {
                rfbsockTCP = AcceptTcpConnection(listenSocket);
                if (rfbsockTCP < 0) {
                    Cleanup(1);
                }

                if (!SetNonBlocking(rfbsockTCP)) {
                    Cleanup(1);
                }

                g_UseUdpSocket = SDL_FALSE;

                // Close the TCP listening socket
                close(listenSocket);
            }
            return;
        }
    }
}
