/*
 *  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

/*
 * misc.c - miscellaneous functions.
 */

#include <signal.h>
#include "vnes.h"
#include <vncviewer.h>
#include "libkeyinput.h"

#ifdef mips
#define USE_ESCVP21
#include "SDL_e21.h"
#else
#undef USE_ESCVP21
#endif

static void CleanupSignalHandler(int sig);

/*
 * ToplevelInitBeforeRealization sets the title, geometry and other resources
 * on the toplevel window.
 */

void
ToplevelInitBeforeRealization()
{

    /* Set the window caption and icon name */
    SDL_WM_SetCaption("sdlvnc", "sdlvnc");

    /* Test if the keyboard is grabbed.  If so, it's probably because the
       XDM login window is up, so try iconifying it to release the grab */

    if (SDL_WM_GrabInput(SDL_GRAB_QUERY) == SDL_GRAB_ON) {
        SDL_WM_GrabInput(SDL_GRAB_OFF);
    }

    /* Set handlers for signals and X errors to perform cleanup */
    signal(SIGHUP, CleanupSignalHandler);
    signal(SIGINT, CleanupSignalHandler);
    signal(SIGTERM, CleanupSignalHandler);
    signal(SIGPIPE, CleanupSignalHandler);
}


/*
 * ToplevelInitAfterRealization initialises things which require the X windows
 * to exist.  It goes into full-screen mode if appropriate, and tells the
 * window manager we accept the "delete window" message.
 */

void
ToplevelInitAfterRealization()
{
    if ( !appData.fullScreen ) {
        FullScreenOff();
    }
}


/*
 * Cleanup - perform any cleanup operations prior to exiting.
 */

void
Cleanup(int code)
{
    Uint32 black;

    signal(SIGHUP, SIG_IGN);
    signal(SIGINT, SIG_IGN);
    signal(SIGTERM, SIG_IGN);
    signal(SIGPIPE, SIG_IGN);
    signal(SIGCHLD, SIG_IGN);

    fprintf(stderr,"Cleanup called\n");

#ifdef BENCHMARK
    fprintf(stderr, "Encrypted Bytes: %d  Encrypt Time: %d ms\n",
            encryptBytes, encryptTime);
    fprintf(stderr, "Decrypted Bytes: %d  Decrypt Time: %d ms\n",
            decryptBytes, decryptTime);
#endif /* BENCHMARK */

#ifdef USE_ESCVP21
    SDL_E21Quit();  // libSDLescvp21.so quit
#endif //USE_ESCVP21

    if ( dpy ) {
        black = SDL_MapRGB(dpy->format, 0,0,0);
        SDL_FillRect(dpy, NULL, black);
        SDL_Flip(dpy);
    }

    SDL_Quit();

    DisconnectDomainSocket();

#ifdef mips
    if (LKI_CloseSocket() != LKI_SUCCESS) {
        exit(-1);
    }
#endif

    fprintf(stderr,"VNeS exiting\n");

    exit(code);
}

static void
CleanupSignalHandler(int sig)
{
    fprintf(stderr,"CleanupSignalHandler called: %d\n", sig);
    Cleanup(1);
}
