/*
 *  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

/*
 * sockets.c - functions to deal with sockets.
 */

#include <unistd.h>
#include <sys/socket.h>
#include <errno.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <assert.h>
#include <vncviewer.h>
#include <sys/time.h>
#include "VNeSCommon.h"
#include "vnes.h"
#include "config.h"

#include "d3des.h"

#ifdef DEBUG
static void PrintInHex(char *buf, int len);
#endif

static SDL_bool WriteExact(int sock, char *buf, int n);

SDL_bool errorMessageOnReadFailure = SDL_TRUE;

// Buffer for UDP packet reading
#define UDP_BUF_SIZE (128*1024)
static int   g_udpSize = 0;
static char  g_udpBuf[UDP_BUF_SIZE];
static char* g_udpPtr = g_udpBuf;

// For encryption
#define DESKEY_SIZE		8

typedef union {
	int size;
	char pad[DESKEY_SIZE];
} des_header_t;

int g_UseEncryption;

// Buffer for decrypted data that can be passed to back 
// to the application
#define BUF_SIZE 65536
static unsigned int g_TotalMessage = 0; // Number of bytes in the current message
static char g_decryptedSize = 0;        // Number of bytes in decryption buffer
static char g_decryptedBuf[BUF_SIZE];   // Buffer for decrypted characters
static char* g_decryptedPtr = g_decryptedBuf;   // Pointer to decryption buffer

#ifdef BENCHMARK
unsigned int decryptTime;
unsigned int decryptBytes;
unsigned int encryptTime;
unsigned int encryptBytes;
#endif

static void
dataEncrypt( char* buf, unsigned int len )
{
	char* cur;

	assert( (len % 8) == 0 );

#ifdef BENCHMARK
	struct timeval start, end;

	encryptBytes += len;
	gettimeofday(&start, NULL);
#endif

	deskey( appData.encryptString, EN0 );

	// (D)encrypt in-place
	cur = buf;
	while ( len > 0 ) {
		des( cur, cur );
		cur += 8;
		len -= 8;
	}

#ifdef BENCHMARK
	gettimeofday(&end, NULL);
	encryptTime += ( end.tv_sec*1000 + end.tv_usec/1000 ) -
		( start.tv_sec*1000 + start.tv_usec/1000 );
#endif	
}

static void
dataDecrypt( char* buf, unsigned int len )
{
	char* cur;

	assert( (len % 8) == 0 );

#ifdef BENCHMARK
	struct timeval start, end;

	decryptBytes += len;
	gettimeofday(&start, NULL);
#endif
	deskey( appData.encryptString, DE1 );

	// (D)encrypt in-place
	cur = buf;
	while ( len > 0 ) {
		des( cur, cur );
		cur += 8;
		len -= 8;
	}

#ifdef BENCHMARK
	gettimeofday(&end, NULL);
	decryptTime += ( end.tv_sec*1000 + end.tv_usec/1000 ) -
		( start.tv_sec*1000 + start.tv_usec/1000 );
#endif
}

/*
 * ReadFromRFBServer is called whenever we want to read some data from the RFB
 * server.  It is non-trivial for two reasons:
 *
 * 1. For efficiency it performs some intelligent buffering, avoiding invoking
 *    the read() system call too often.  For small chunks of data, it simply
 *    copies the data out of an internal buffer.  For large amounts of data it
 *    reads directly into the buffer provided by the caller.
 *
 * 2. Whenever read() would block, it invokes the Xt event dispatching
 *    mechanism to process X events.  In fact, this is the only place these
 *    events are processed, as there is no XtAppMainLoop in the program.
 */

static SDL_bool rfbsockReady = SDL_FALSE;
static void
ProcessEvents()
{
    SDL_Event event;
    struct timeval tm;
    int fd;
    int ret;
    fd_set   fds;

    if ( g_UseUdpSocket ) {
        fd = rfbsockUDP;
    } else {
        fd = rfbsockTCP;
    }

	rfbsockReady = SDL_FALSE;
    while (rfbsockReady == SDL_FALSE) {
        /* Poll the dialog to update its status */
        PollDialog(NULL);

        if (SDL_PollEvent(&event)) {
            if (event.type == SDL_VIDEOEXPOSE || 
                event.type == SDL_ACTIVEEVENT) 
            {
                HandleBasicDesktopEvent(&event);
            } else {
                /* Send the event to the dialog */
                PollDialog(&event);
                SendRFBEvent(&event, NULL, NULL);

                if ( event.type == KEY_EVENT_TYPE ) {
                    SDL_Event repeat;
                    Uint32 starttime = SDL_GetTicks();

                    // Throw out any key events for the next 20 ms
                    while(SDL_PeepEvents(&repeat, 1, SDL_GETEVENT, 
                                         (SDL_KEYDOWNMASK|SDL_KEYUPMASK)) > 0)
                    {
                        if ((SDL_GetTicks() - starttime) > 20) {
                            // break after 20 millisec of events to
                            // avoid network starvation via key input
                            break;
                        }
                    }
                }
            }
        }

        FD_ZERO(&fds);
        FD_SET(fd,&fds);
        tm.tv_usec = 20000; /* 20 millisec timeout */
        tm.tv_sec = 0;

        ret = select(fd+1, &fds, NULL, NULL, &tm);
        if (ret > 0) {
            rfbsockReady = SDL_TRUE;
        } 
    }
}

static int
ReadExactUdp( int sock, char* out, unsigned int len )
{
    unsigned int count = 0;
    int i;

    // Read another UDP packet into the buffer if necessary
    while( g_udpSize == 0 ) {
        assert( g_udpPtr == g_udpBuf );

        i = read(sock, g_udpPtr, UDP_BUF_SIZE);
        if (i <= 0) {
            if (i < 0) {
                if (errno == EWOULDBLOCK || errno == EAGAIN) {

                    ProcessEvents();
                    i = 0;

                } else if (errno == EMSGSIZE) {
                        
                    fprintf(stderr,"UDP packet larger than buffer.\n");
                    return -1;

                } else {
                    fprintf(stderr,programName);
                    perror(": read");
                    return -1;
                }
            } else {

                ProcessEvents();
            }
        }else {
            g_udpSize = i;
            g_udpPtr = g_udpBuf;

#ifndef NDEBUG
            // PrintInHex(g_udpPtr, i);
#endif            
        }
    }

    // Read remaining packet data from the UDP buffer
    if ( len <= g_udpSize ) {
        memcpy(out, g_udpPtr, len);
        g_udpSize -= len;
        g_udpPtr += len;
        count += len;
    } else {
        fprintf(stderr, "RFB message is bigger than a single UDP packet!\n");
        return -1;
    }

    return count;
}

static int
ReadExact( int sock, char* out, unsigned int len )
{
	int count = 0;
    int i;

    if ( g_UseUdpSocket ) {
        return ReadExactUdp( sock, out, len );
    }

    // Stating TCP read
	while (count < len) {
		i = read(sock, out + count, len - count);
		if (i <= 0) {
			if (i < 0) {
				if (errno == EWOULDBLOCK || errno == EAGAIN) {
					ProcessEvents();
					i = 0;
				} else {
					fprintf(stderr,programName);
					perror(": read");
					return -1;
				}
			} else {
				if (errorMessageOnReadFailure) {
					fprintf(stderr,"%s: VNC server closed connection\n",programName);
				}
				return -1;
			}
		}else {
#ifndef NDEBUG
            // PrintInHex(out+count,i);
#endif

			count += i;
		}
	}

#ifdef DEBUG
	fprintf(stderr, "ReadExact: TCP = %d\n", count );
#endif

	return count;
}

SDL_bool
ReadDecryptHeader( int sock )
{
	des_header_t hdr;
	int count;

	assert( g_TotalMessage == 0 );

	// We are expecting the 8 byte DES encrypted size header
	count = ReadExact( sock, (char*)&hdr, DESKEY_SIZE );
	if ( count == -1 ) {
		return SDL_FALSE;
	}

	// Decrypt data
	dataDecrypt( (char*)&hdr, count );
	g_TotalMessage = Swap32IfLE(hdr.size);

#ifdef DEBUG
	fprintf(stderr, "Got DES header size: %d\n", g_TotalMessage );
#endif

	assert( g_TotalMessage > 0 );
	return SDL_TRUE;
}

SDL_bool
WriteEncryptHeader( int sock, unsigned int len )
{
	des_header_t hdr;

	assert( len > 0 );

	// Encrypt data
	hdr.size = Swap32IfLE(len);
	dataEncrypt( (char*)&hdr, sizeof(hdr) );

#ifdef DEBUG
	fprintf(stderr, "Sent DES header size: %d\n", len );
#endif
	
	// We are expecting the 8 byte DES encrypted size header
	return WriteExact(sock, (char*)&hdr, sizeof(hdr));
}

int
ReadDecryptBody( int sock, char *out, unsigned int n )
{
	int total = 0;
	int count;
	int size;

	assert( g_TotalMessage > 0 );
	assert( n <= g_TotalMessage );
	assert( g_decryptedSize == 0 );

	// Round n down to nearest multiple of 8 (for decrypt)
	size = (n / DESKEY_SIZE) * DESKEY_SIZE;
	if ( size > 0 ) {
		count = ReadExact( sock, out, size );
		if ( count == -1 ) {
			return -1;
		}

		// Decrypt data
		dataDecrypt( out, count );
		out += count;
		total += count;

		assert( g_TotalMessage >= count );
		g_TotalMessage -= count;
	}

	if ( (n % DESKEY_SIZE) > 0 ) {
		int remainder = n % DESKEY_SIZE;

		// Read a partial block of data
		count = ReadExact( sock, g_decryptedBuf, DESKEY_SIZE );
		if ( count == -1 ) {
			return -1;
		}

		// Decrypt data
		dataDecrypt( g_decryptedBuf, count );
		memcpy(out, g_decryptedBuf, remainder);
		total += remainder;

		assert( g_TotalMessage >= remainder );
		g_TotalMessage -= remainder;

		if ( g_TotalMessage > 0 ) {
			g_decryptedPtr = g_decryptedBuf + remainder;
			g_decryptedSize = DESKEY_SIZE - remainder;
			if ( g_decryptedSize > g_TotalMessage ) {
				g_decryptedSize = g_TotalMessage;
			}
			g_TotalMessage -= g_decryptedSize;
		}
	}

	return total;
}

SDL_bool
DecryptFromRFBServer( int sock, char *out, unsigned int n )
{
	int ret;

	// The are enough decrypted bytes already
	if ( n <= g_decryptedSize ) {
		memcpy( out, g_decryptedPtr, n );
		g_decryptedSize -= n;
		g_decryptedPtr += n;
		return SDL_TRUE;
	}

	if ( g_decryptedSize > 0 ) {
		// Empty any remaining partially-decrypted blocks
		memcpy( out, g_decryptedPtr, g_decryptedSize );
		n -= g_decryptedSize;	
		out += g_decryptedSize;	
		g_decryptedPtr = g_decryptedBuf;
		g_decryptedSize = 0;
	}

	// Update the g_TotalMessage variable if necessary
	if ( g_TotalMessage == 0 ) {
		if ( ReadDecryptHeader(sock) == SDL_FALSE ) {
			return SDL_FALSE;
		}
	}

    // Repeat until there is sufficient data available 
    // in a single packet to fulfill 'n' bytes
	while (n > g_TotalMessage) {
		// Read a message body
		ret = ReadDecryptBody( sock, out, g_TotalMessage );
		if ( ret == -1 ) {
			return SDL_FALSE;
		}
		n -= ret;
		out += ret;

		// Update the g_TotalMessage variable if necessary
		if ( g_TotalMessage == 0 ) {
			if ( ReadDecryptHeader(sock) == SDL_FALSE ) {
				return SDL_FALSE;
			}
		}
	}

    // Read remaining required data
	if ( n > 0 ) {
		// Read a message body
		return (ReadDecryptBody( sock, out, n ) != -1);
	}

	return SDL_TRUE;
}

SDL_bool
ReadFromRFBServer( char *out, unsigned int n)
{
	int sock = g_UseUdpSocket ? rfbsockUDP : rfbsockTCP;

	if ( g_UseEncryption ) {
		return DecryptFromRFBServer( sock, out, n );
	}

    if (n <= g_decryptedSize) {
		memcpy(out, g_decryptedPtr, n);
		g_decryptedPtr += n;
		g_decryptedSize -= n;
		return SDL_TRUE;
    }

    memcpy(out, g_decryptedPtr, g_decryptedSize);

    out += g_decryptedSize;
    n -= g_decryptedSize;

    g_decryptedPtr = g_decryptedBuf;
    g_decryptedSize = 0;

    return (ReadExact( sock, out, n ) != -1 );
}


/*
 * Write an exact number of bytes, and don't return until you've sent them.
 */

SDL_bool
WriteExact(int sock, char *buf, int n)
{
	fd_set fds;
	int i = 0;
	int j;

	while (i < n) {
		j = write(sock, buf + i, (n - i));
		if (j <= 0) {
			if (j < 0) {
				if (errno == EWOULDBLOCK || errno == EAGAIN) {
					FD_ZERO(&fds);
					FD_SET(rfbsockTCP,&fds);

					if (select(rfbsockTCP+1, NULL, &fds, NULL, NULL) <= 0) {
						fprintf(stderr,programName);
						perror(": select");
						return SDL_FALSE;
					}
					j = 0;
				} else {
					fprintf(stderr,programName);
					perror(": write");
					return SDL_FALSE;
				}
			} else {
				fprintf(stderr,"%s: write failed\n",programName);
				return SDL_FALSE;
			}
		}
		i += j;
	}

#ifdef DEBUG
	fprintf( stderr, "WriteExact: Sent %d bytes\n", i );
#endif

	return SDL_TRUE;
}

SDL_bool
EncryptToRFBServer( int sock, char* out, unsigned int n )
{
	int size;

	assert( n > 0 );

	if ( WriteEncryptHeader( sock, n ) == SDL_FALSE ) {
		return SDL_FALSE;
	}

	size = (n / DESKEY_SIZE) * DESKEY_SIZE;
	if ( size > 0 ) {
		dataEncrypt( out, size );
		if ( WriteExact( sock, out, size ) == SDL_FALSE ) {
			return SDL_FALSE;
		}
		out += size;
		n -= size;
	}

	if ( (n % DESKEY_SIZE) > 0 ) {
		char encrypt[DESKEY_SIZE] = {0,0,0,0,0,0,0,0};

		memcpy( encrypt, out, (n % DESKEY_SIZE) );
		dataEncrypt( encrypt, DESKEY_SIZE );

		if ( WriteExact( sock, encrypt, DESKEY_SIZE ) == SDL_FALSE ) {
			return SDL_FALSE;
		}
	}

	return SDL_TRUE;
}

SDL_bool
WriteToRFBServer( char *out, unsigned int n)
{
#if CAN_SEND
	if ( g_UseEncryption ) {
		return EncryptToRFBServer( rfbsockTCP, out, n );
	} else {
		return WriteExact( rfbsockTCP, out, n );
	}
#else // CAN_SEND
    return SDL_TRUE;
#endif // CAN_SEND
}


/*
 * ConnectToTcpAddr connects to the given TCP port.
 */

int
ConnectToTcpAddr(unsigned int host, int port)
{
    int sock;
    struct sockaddr_in addr;
    int one = 1;

    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = host;

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
        fprintf(stderr,programName);
        perror(": ConnectToTcpAddr: socket");
        return -1;
    }

    if (connect(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
        fprintf(stderr,programName);
        perror(": ConnectToTcpAddr: connect");
        close(sock);
        return -1;
    }

    if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY,
                   (char *)&one, sizeof(one)) < 0) {
        fprintf(stderr,programName);
        perror(": ConnectToTcpAddr: setsockopt");
        close(sock);
        return -1;
    }

    return sock;
}


/*
 * Create a UDP socket and listen on the requested port in broadcast
 * mode
 */

int
CreateDataSocket(int port)
{
	int sock;
	struct sockaddr_in addr;
	int one = 1;

	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr.s_addr = INADDR_ANY;

	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock < 0) {
		fprintf(stderr,programName);
		perror(": CreateDataSocket: socket");
		return -1;
	}

	if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		fprintf(stderr,programName);
		perror(": CreateDataSocket: bind");
		close(sock);
		return -1;
	}
#ifdef DEBUG
	fprintf(stderr, "Bound Data socket to port: %d\n", port);
#endif

	if (setsockopt(sock, SOL_SOCKET, SO_BROADCAST, &one, sizeof(one)) <0) {
		fprintf(stderr,programName);
		perror(": CreateDataSocket: setsockopt");
		close(sock);
		return -1;
	}

	if ( SetNonBlocking(sock) == SDL_FALSE ) {
		fprintf(stderr,programName);
		perror(": CreateDataSocket: SetNonBlocking");
		close(sock);
		return -1;
	}

	return sock;
}

/*
 * FindFreeTcpPort tries to find unused TCP port in the range
 * (TUNNEL_PORT_OFFSET, TUNNEL_PORT_OFFSET + 99]. Returns 0 on failure.
 */

int
FindFreeTcpPort(void)
{
    int sock, port;
    struct sockaddr_in addr;

    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = INADDR_ANY;

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
        fprintf(stderr,programName);
        perror(": FindFreeTcpPort: socket");
        return 0;
    }

    for (port = TUNNEL_PORT_OFFSET + 99; port > TUNNEL_PORT_OFFSET; port--) {
        addr.sin_port = htons((unsigned short)port);
        if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) == 0) {
            close(sock);
            return port;
        }
    }

    close(sock);
    return 0;
}


/*
 * ListenAtTcpPort starts listening at the given TCP port.
 */

int
ListenAtTcpPort(int port)
{
    int sock;
    struct sockaddr_in addr;
    int one = 1;

    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = INADDR_ANY;

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
        fprintf(stderr,programName);
        perror(": ListenAtTcpPort: socket");
        return -1;
    }

    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR,
                   (const char *)&one, sizeof(one)) < 0) {
        fprintf(stderr,programName);
        perror(": ListenAtTcpPort: setsockopt");
        close(sock);
        return -1;
    }

    if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
        fprintf(stderr,programName);
        perror(": ListenAtTcpPort: bind");
        close(sock);
        return -1;
    }

    if (listen(sock, 5) < 0) {
        fprintf(stderr,programName);
        perror(": ListenAtTcpPort: listen");
        close(sock);
        return -1;
    }

    return sock;
}


/*
 * ListenAtTcpPort starts listening at the given TCP port.
 */

int
ListenAtUdpPort(int port)
{
    int sock;
    struct sockaddr_in addr;
    int one = 1;

    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = INADDR_ANY;

    sock = socket(AF_INET, SOCK_DGRAM, 0);
    if (sock < 0) {
        fprintf(stderr,programName);
        perror(": ListenAtTcpPort: socket");
        return -1;
    }

    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR,
                   (const char *)&one, sizeof(one)) < 0) {
        fprintf(stderr,programName);
        perror(": ListenAtTcpPort: setsockopt");
        close(sock);
        return -1;
    }

    if (setsockopt(sock, SOL_SOCKET, SO_BROADCAST,
                   (const char *)&one, sizeof(one)) < 0) {
        fprintf(stderr,programName);
        perror(": ListenAtTcpPort: setsockopt");
        close(sock);
        return -1;
    }

    if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
        fprintf(stderr,programName);
        perror(": ListenAtTcpPort: bind");
        close(sock);
        return -1;
    }

    return sock;
}


/*
 * AcceptTcpConnection accepts a TCP connection.
 */

int
AcceptTcpConnection(int listenSock)
{
    int sock;
    struct sockaddr_in addr;
    int addrlen = sizeof(addr);
    int one = 1;

    sock = accept(listenSock, (struct sockaddr *) &addr, &addrlen);
    if (sock < 0) {
        fprintf(stderr,programName);
        perror(": AcceptTcpConnection: accept");
        return -1;
    }

    if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY,
                   (char *)&one, sizeof(one)) < 0) {
        fprintf(stderr,programName);
        perror(": AcceptTcpConnection: setsockopt");
        close(sock);
        return -1;
    }

    return sock;
}


/*
 * SetNonBlocking sets a socket into non-blocking mode.
 */

SDL_bool
SetNonBlocking(int sock)
{
    if (fcntl(sock, F_SETFL, O_NONBLOCK) < 0) {
        fprintf(stderr,programName);
        perror(": SetNonBlocking: fcntl");
        return SDL_FALSE;
    }
    return SDL_TRUE;
}


/*
 * StringToIPAddr - convert a host string to an IP address.
 */

SDL_bool
StringToIPAddr(const char *str, unsigned int *addr)
{
    struct hostent *hp;

    if (strcmp(str,"") == 0) {
        *addr = 0; /* local */
        return SDL_TRUE;
    }

    *addr = inet_addr(str);

    if (*addr != -1)
        return SDL_TRUE;

    hp = gethostbyname(str);

    if (hp) {
        *addr = *(unsigned int *)hp->h_addr;
        return SDL_TRUE;
    }

    return SDL_FALSE;
}


/*
 * Test if the other end of a socket is on the same machine.
 */

SDL_bool
SameMachine(int sock)
{
    struct sockaddr_in peeraddr, myaddr;
    int addrlen = sizeof(struct sockaddr_in);

    getpeername(sock, (struct sockaddr *)&peeraddr, &addrlen);
    getsockname(sock, (struct sockaddr *)&myaddr, &addrlen);

    return (peeraddr.sin_addr.s_addr == myaddr.sin_addr.s_addr);
}


#ifdef DEBUG
/*
 * Print out the contents of a packet for debugging.
 */

void
PrintInHex(char *buf, int len)
{
    int i, j;
    char c, str[17];

    str[16] = 0;

    fprintf(stderr,"ReadExact: ");

    for (i = 0; i < len; i++)
    {
        if ((i % 16 == 0) && (i != 0)) {
            fprintf(stderr,"           ");
        }
        c = buf[i];
        str[i % 16] = (((c > 31) && (c < 127)) ? c : '.');
        fprintf(stderr,"%02x ",(unsigned char)c);
        if ((i % 4) == 3)
            fprintf(stderr," ");
        if ((i % 16) == 15)
        {
            fprintf(stderr,"%s\n",str);
        }
    }
    if ((i % 16) != 0)
    {
        for (j = i % 16; j < 16; j++)
        {
            fprintf(stderr,"   ");
            if ((j % 4) == 3) fprintf(stderr," ");
        }
        str[i % 16] = 0;
        fprintf(stderr,"%s\n",str);
    }

    fflush(stderr);
}
#endif


/*!  Read the VNeS header and verify that the fields match our
  settings.

  \return -1 on read error, 0 on mismatch or 1 on okay packet.
*/
int
ReadVnesHeader()
{
    static SDL_bool isFullscreen = SDL_FALSE;
    VNESUDPHEADER msgHeader;
    int size;
    int bytes;
    int sock;

    // fprintf(stderr, "Reading VNeS header...\n");

    // Want to make sure we process events at least once per update
    ProcessEvents();

    // Update the UDP buffer's counters
    if ( g_UseUdpSocket ) {
        if ( g_udpSize != 0 ) {
            fprintf(stderr, "Left over data from last UDP packet: %d\n", 
                    g_udpSize);
        }
        g_udpSize = 0;
        g_udpPtr = g_udpBuf;
    }

    // Use the appropriate socket and read buffer    
    if ( g_UseUdpSocket ) {
        sock = rfbsockUDP;
        size = sizeof(VNESUDPHEADER);
    } else {
        sock = rfbsockTCP;
        size = sizeof(VNESHEADER);
    }

    bytes = ReadExact(sock, (char*)&msgHeader, size);
    if ( bytes != size ) {
        fprintf(stderr, "Incorrect number of bytes read. Expected: %d  Got: %d\n",
                size, bytes);
        return -1;
    }

    // Verify the magic number is intact
    if (memcmp(msgHeader.vnesHeader.dwMagicNum, VNES_MAGICNUMBER, 4)) {
        fprintf(stderr, "VNeS magic number mismatch\n");
        return 0;
    }

    // Verify the major version is okay
    if (memcmp(msgHeader.vnesHeader.MajorVersion, VNES_MAJORVERSION, 2)) {
        fprintf(stderr, "VNeS major version mismatch\n");
        return 0;
    }

    // Verify the server address matches ours
    if (memcmp(msgHeader.vnesHeader.dwServerIP, appData.serverIP, 4)) {
        fprintf(stderr, "VNeS server mismatch\n");
        return 0;
    }

    if (g_UseUdpSocket ) {
        msgHeader.dwCommandID = Swap32IfLE(msgHeader.dwCommandID);
        msgHeader.dwDataID = Swap32IfLE(msgHeader.dwDataID);
        msgHeader.dwFullScrID = Swap32IfLE(msgHeader.dwFullScrID);
        msgHeader.dwFullScrTotal = Swap32IfLE(msgHeader.dwFullScrTotal);

#ifdef DEBUG
        fprintf(stderr, "Data ID: %ld\n", msgHeader.dwDataID);
#endif

        if ( msgHeader.dwFullScrID != 0 ) {
            if ( isFullscreen == SDL_FALSE ) {
                // Got the first packet of a fullscreen update
                StartFullscreenUpdate(msgHeader.dwFullScrTotal);
                isFullscreen = SDL_TRUE;
            }

#ifndef NDEBUG            
            fprintf(stderr, "Fullscreen update: %ld of %ld\n", 
                    msgHeader.dwFullScrID, msgHeader.dwFullScrTotal);
#endif

            // Acknowledge the packet
            AckFullscreenUpdate(msgHeader.dwFullScrID);

            // If we got the last update, finish up
            if ( msgHeader.dwFullScrID == msgHeader.dwFullScrTotal ) {
                EndFullscreenUpdate();
                isFullscreen = SDL_FALSE;
            }
        } else {
            if ( isFullscreen == SDL_TRUE ) {
                //EndFullscreenUpdate();
                isFullscreen = SDL_FALSE;
            }            
        }
    }

    return 1;
}
