#include <cassert>
#include "vncviewer.h"
#include "config.h"
#include "debug.h"
#include "helper.h"
#include "Dialog.h"

#ifdef mips
#include "libkeyinput.h"
#define USE_ESCVP21
#include "SDL_e21.h"
#else
#undef USE_ESCVP21
#endif

const Uint32 DIALOG_TIMEOUT = 7000;      // 7 sec timeout

static Dialog* sg_dialog = 0;
static Uint32  sg_timeout;

static void EventUser(const SDL_Event* event);

void InitUI() 
{
    Uint8 lang;
#ifdef mips
    if (LKI_OpenSocket(NULL) != LKI_SUCCESS) {
        DebugPrint("***** Failure to initialize LKI\n");
        Cleanup(EXIT_FAILURE);
    }
    lang = get_lang();
#else
    lang = 0x01;
#endif  

    if (UIRsrcMgr::Instance(lang)->Init(UIRsrcMgr::ModVNC)) {
        DebugPrint("***** Failure to initialize UI resource manager\n");
        Cleanup(EXIT_FAILURE);
    }
}

void OpenDialog()
{
    DebugPrint("%s: Opening Dialog\n", __FUNCTION__);

    assert( sg_dialog == 0 );

    Uint8 lang;
#ifdef mips
    if (LKI_OpenSocket(NULL) != LKI_SUCCESS) {
        DebugPrint("***** Failure to initialize LKI\n");
        Cleanup(EXIT_FAILURE);
    }
    lang = get_lang();
#else
    lang = 0x01;
#endif  
    UIRsrcMgr::Instance()->SetLang(lang);

    SDL_Surface* screen = SDL_GetVideoSurface();
    SDL_Rect rect = VNES_DIALOG_DEFAULT_RECT;
    rect.w = screen->w;
    rect.y = screen->h - rect.h;
    sg_dialog = new Dialog(screen, rect);

    sg_dialog->Init();
}


void DrawDialog()
{
    if ( sg_dialog == 0 ) {
        return;
    }

    sg_dialog->Draw();
    sg_dialog->Update();
}

void PollDialog(const SDL_Event* event)
{
    // If there is no existing dialog window and the ESC key was
    // pressed then create a new one and return.  For any other event,
    // just return.
    if ( event ) {
        if ( event->type == KEY_EVENT_TYPE ) {
            DebugPrint("%s: Got usuable key event\n", __FUNCTION__);
            sg_timeout = SDL_GetTicks();

            if ( sg_dialog == 0 ) {
                if ( event->key.keysym.sym == SDLK_ESCAPE ) {        
                    OpenDialog();
                    DrawDialog();
                }
                return;
            }
        } else if ( event->type == SDL_USEREVENT ) {
            DebugPrint("%s: Got user event\n", __FUNCTION__);
            EventUser(event);
        }
    }

    if ( sg_dialog ) {
        if ((SDL_GetTicks() - sg_timeout) > DIALOG_TIMEOUT) {
            DebugPrint("%s: Timeout!\n", __FUNCTION__);
            CloseDialog();
            return;
        }
        DebugPrint("%s: Timeout in: %d ms\n", __FUNCTION__,
                   DIALOG_TIMEOUT - (SDL_GetTicks()-sg_timeout));
    }

    // Pass the event on to the dialog handler
    if ( sg_dialog && event ) {
        DebugPrint("%s: Passing event type '%d' to dialog\n", 
                   __FUNCTION__, (int)event->type);
        sg_dialog->Event(*event);
    }
}

void CloseDialog()
{
    DebugPrint("%s: Closing Dialog\n", __FUNCTION__);
    assert( sg_dialog != 0 );
    delete sg_dialog;
    sg_dialog = 0;
    sg_timeout = 0;
}


static void
EventUser(const SDL_Event* event)
{
#ifdef USE_ESCVP21
    assert( event != 0 );

    DebugPrint("EventUser\n");
    if (event->user.code != SDL_E21_MAGIC) {
        return;
    }

    if (!event->user.data1) {
        return;
    }

    SDL_E21Event *e21ev = static_cast<SDL_E21Event *>(event->user.data1);
    switch(e21ev->code) {
    case EV_LANG:
        UIRsrcMgr::Instance()->SetLang(e21ev->param[0]);

        if ( sg_dialog ) {
            sg_dialog->UpdateResources();
            sg_dialog->Redraw();
            sg_dialog->Update();
        }

		LKI_Send(cmdIMSTAT, "21");
		break;
    default:
        break;
    }

    free(event->user.data1);
    const_cast<SDL_Event*>(event)->user.data1 = NULL;
#endif
}
