/*
 *  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

/*
 * vncviewer.c - the Xt-based VNC viewer.
 */

#include <stdlib.h>
#include "SDL_e21.h"
#include "vnes.h"
#include "vncviewer.h"

char *programName;
SDL_Surface* dpy;

int
main(int argc, char **argv)
{
    int i;
    int flags;
#ifdef UI_TEST
    Uint32 colour;
    SDL_Event event;
#endif

    dpy = NULL;
    programName = argv[0];
    
    /* Initialize the UI resource manager */
    InitUI();

#ifndef UI_TEST
    /* Start the domain socket */
    if (ConnectDomainSocket()) {        
        Cleanup(1);
    }
#endif

    /* Initialize the SDL video and event subsystems */
    flags = SDL_INIT_VIDEO|SDL_INIT_NOPARACHUTE;
    SDL_Init(flags);

#ifdef USE_ESCVP21
    // initialize libSDLescvp.so 
    if(SDL_E21Init() < 0){
        fprintf(stderr, "Could not initialize ESCVP21\n");
        Cleanup(1);
    }
#endif //USE_ESCVP21

#ifdef UI_TEST
    flags = SDL_SWSURFACE;
    dpy = SDL_SetVideoMode(1024, 768, 0, flags);

    /* Fill the screen with a nice blue */
    colour = SDL_MapRGB(dpy->format, 65, 105, 225);
    SDL_FillRect(dpy, NULL, colour);
    SDL_Flip(dpy);

    if ( dpy == 0 ) {
        fprintf(stderr, "Could not initialize video display\n");
        Cleanup(1);
    }

    while(1) {
        PollDialog(NULL);

        if (SDL_PollEvent(&event)) {
            PollDialog(&event);
        }
        usleep(50);
    }
#endif

    /* The -listen option is used to make us a daemon process which listens for
       incoming connections from servers, rather than actively connecting to a
       given server. The -tunnel and -via options are useful to create
       connections tunneled via SSH port forwarding. We must test for the
       -listen option before invoking any Xt functions - this is because we use
       forking, and Xt doesn't seem to cope with forking very well. For -listen
       option, when a successful incoming connection has been accepted,
       listenForIncomingConnections() returns, setting the listenSpecified
       flag. */

    for (i = 1; i < argc; i++) {
        if (strcmp(argv[i], "-listen") == 0) {
            listenForIncomingConnections(&argc, argv, i);
            break;
        }
    }

    /* Interpret resource specs and process any remaining command-line
       arguments (i.e. the VNC server name).  If the server name isn't
       specified on the command line, getArgsAndResources() will pop
       up a dialog box and wait for one to be entered. */
    GetArgsAndResources(argc, argv);

    /* Unless we accepted an incoming connection, make a TCP
       connection to the given VNC server */
    if (!listenSpecified) {
        if (!ConnectToRFBServer(vncServerHost, vncServerPort)) {
            Cleanup(1);
        }
    }

    /* Initialise the VNC connection, including reading the
       password */
    if (!InitialiseRFBConnection()) {
        Cleanup(1);
    }

    /* Create the "desktop" widget, and perform initialisation which
       needs doing before the widgets are realized */

    ToplevelInitBeforeRealization();

    DesktopInitBeforeRealization();

    /* Find the best pixel format and X visual/colormap to use */
    SetVisualAndCmap();

    /* Perform initialisation that needs doing after realization, now
       that the X windows exist */
    ToplevelInitAfterRealization();

    DesktopInitAfterRealization();

    /* Tell the VNC server which pixel format and encodings we want to
       use */
    SetFormatAndEncodings();

    /* Now enter the main loop, processing VNC messages.  X events
       will automatically be processed whenever the VNC connection is
       idle. */
    if ( appData.useUdpSocket ) {
        g_UseUdpSocket = SDL_TRUE;
    }

    while (1) {
        if (!HandleRFBServerMessage())
            break;
    }

    Cleanup(0);
    return 0;
}
