/*!
  \file   vnes.cpp
  \author Aric Cyr
  \date   Thu Jul 18 10:59:05 2002
  
  \brief  VNeS server communication protocol implementation
  
  
*/

#include <cassert>
#include <cstring>
#include <deque>
#include "debug.h"
#include "domain.h"
#include "netutil.h"
#include "vnesproto.h"
#include "vnes.h"

#define POLL_TIMEOUT    10   // 10 millisec
#define WINDOW_SIZE     5    // 5 packet window size

// type definitions
typedef std::deque<unsigned int> RequestQueue;

// forward function declarations
static bool SendRequest(unsigned int itemID);

static RequestQueue s_requests;
static Domain* s_servSocket;

/*!  Connect to the EasyConnect daemon via a unix domain socket.

  \return 0 on success, otherwise error.
*/
int ConnectDomainSocket()
{
    DebugPrint("Entering: %s\n", __func__);

    if ( s_servSocket != 0 ) {
        DebugPrint("%s called on already connected socket.\n", __func__);
        delete s_servSocket;
        s_servSocket = 0;
    }

    s_servSocket = new Domain();
    if ( s_servSocket == 0 ) {
        return -1;
    }

    // Initialize the anonymous unix domain path.  The first byte
    // should be NULL so that no file is required in the filesystem.
    char path[UNIX_PATH_MAX];
    memset(path, 0, sizeof(path));
    strcpy(path+1, VNES_SOCKET_PATH);

    if ( ! s_servSocket->Create(Domain::DATAGRAM) ) {
        NetworkUtil::PrintSocketError(*s_servSocket);
        delete s_servSocket;
        s_servSocket = 0;
        return -1;
    }

    if ( ! s_servSocket->Connect(path) ) {
        NetworkUtil::PrintSocketError(*s_servSocket);
        delete s_servSocket;
        s_servSocket = 0;
        return -1;
    }

    s_servSocket->SetTimeout(POLL_TIMEOUT);

    return 0;
}

/*!  Connect to the EasyConnect daemon via a unix domain socket.

  \return 0 on success, otherwise error.
*/
void DisconnectDomainSocket()
{
    DebugPrint("Entering: %s\n", __func__);

    delete s_servSocket;
    s_servSocket = 0;
}

/*!  Check the socket for any pending messages and process them.

  \return 0 on success, otherwise error.
*/
void PollDomainSocket()
{
    // For now do nothing! I don't expect any messages to come this
    // way for now, anyways.
}


/*!  Notification of the start of a new fullscreen update cycle has
begun.  The algorithm we use is to add all ID numbers to the queue and
then remove them one at a time as they are ack'ed.  The queue must
remain in non-decreasing order based on ACK values.  This should be
automatic, since no insertions are done in the queue after
initialization.
  
  \param totalID The total number of expected update packets.
*/
void StartFullscreenUpdate(unsigned int totalID)
{
    DebugPrint("Entering: %s\n", __func__);

    // Add all the possible IDs
    for ( unsigned int i = 1; i <= totalID; i++ ) {
        s_requests.push_back(i);
    }
}


/*!  Acknowledge receipt of a fullscreen update packet.

  \param itemID ID of the received packet.
*/
void AckFullscreenUpdate(unsigned int ackID)
{
    unsigned int minACK;

    DebugPrint("Entering: %s\n", __func__);

    // ACK ids are one-based not zero-based, and since we are using
    // unsigned values, we have to make sure that we don't wrap around
    // to a very big number.
    if ( (WINDOW_SIZE + 1) > ackID ) {
        minACK = 1;
    } else {
        minACK = ackID - WINDOW_SIZE + 1;
    }

    RequestQueue::iterator item = s_requests.begin();
    while ( item != s_requests.end() ) {
        if ((*item) > ackID) {
            // Fast out.  Since the item in the queue has a higher
            // number than the ACK'ed packet, we must have the same
            // packet twice.  Just ignore the ack.
            break;

        } else if ((*item) < minACK) {
            // The item in the queue is outside the acceptable window
            // range, so send it as a request and remove it from the
            // queue.
            SendRequest((*item));
            s_requests.erase(item);

            // Reset the iterator to the first element
            item = s_requests.begin();

        } else if ((*item) < ackID) {
            // We are within the window, so we can just ignore this
            // item and advanced to the next for now.
            item++;

        } else if ((*item) == ackID) {
            // Found our desired item.  Accept the ack, and remove
            // this item from the queue. Then we are done this ack.
            s_requests.erase(item);
            break;

        } else {
            // Strange (and probably impossible) situation.
            DebugPrint("AckFullscreenUpdate: Unhandled possibility occured\n");
            break;
        }
    }
}


/*! Process all queued missing fullscreen updates and send a request
  to have them resent.
*/
void EndFullscreenUpdate()
{
    DebugPrint("Entering: %s\n", __func__);

    while( ! s_requests.empty() ) {
        SendRequest(s_requests.front());
        s_requests.pop_front();
    }
}

/*! Send a message to the controlling daemon.
*/
int SendDomainMessage(const vnes_msg_t* message)
{
    DebugPrint("Entering: %s\n", __FUNCTION__);
    assert( s_servSocket != 0 );
    assert( message != 0 );

    DebugPrint("SendMessage: Sending message ID %d\n", message->type);

    int ret;
    ret = s_servSocket->Send((char*)message, sizeof(vnes_msg_t));
    if ( ret != sizeof(vnes_msg_t) ) {
        if ( ret < 0 ) {
            NetworkUtil::PrintSocketError(*s_servSocket);
        }
        DebugPrint("SendRequest: Failed to send\n");
        return false;
    }

    return true;    
}

static bool
SendRequest(unsigned int itemID)
{
    DebugPrint("Entering: %s\n", __func__);
    assert( s_servSocket != 0 );

    DebugPrint("SendRequest: Sending ID %d\n", itemID);

    vnes_msg_t message;
    message.type = VNES_MSG_REQ_FULLSCREEN;
    message.msgdata.fullscrn.id = itemID;

    int ret;
    ret = s_servSocket->Send((char*)&message, sizeof(message));
    if ( ret != sizeof(message) ) {
        if ( ret < 0 ) {
            NetworkUtil::PrintSocketError(*s_servSocket);
        }
        DebugPrint("SendRequest: Failed to send\n");
        return false;
    }

    return true;
}
