/*
    Copyright (C) SEIKO EPSON CORPORATION 2002-2003. All Rights Reserved.

    3-3-5, Owa, Suwa-shi Nagano-ken, 392-8502 Japan
    Phone: +81-266-52-3131

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/




#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <cstdlib>
#include <cassert>
#include <cstring>

#include "debug.h"
#include "configfile.h"

static const char* option_names[CONFIG_OPTION_LAST] = {
    "keyword",
    "channel",
    "essid",
    "adhocmode",
    "wep",
    "wepmode",
    "wepkey",
    "projectorname",
    "network",
    "dhcp",
    "ip",
    "mask",
    "gateway",
    "password",
    "trap1",
    "trap2",
    "device"
};

ConfigFile::ConfigFile(const char* filename)
{
    if ( filename == 0 ) {
        
        return;
    }

    
    struct stat fileStat;
    if ( stat(filename, &fileStat) == 0 ) {
        m_FilePtr = fopen(filename, "r+");
    } else {
        m_FilePtr = fopen(filename, "w+");
    }

    assert(m_FilePtr && "Could not open config file");

    
    Read();
}

ConfigFile::~ConfigFile()
{
    if ( m_FilePtr ) {
        fclose(m_FilePtr);
    }
}


void
ConfigFile::Read()
{
    char lineBuffer[MaxOptionLength];
    char* optionName;
    char* optionValue;

    
    rewind(m_FilePtr);

    DeleteAllOptions();

    while ( !feof(m_FilePtr) ) {
        
        fgets(lineBuffer, MaxOptionLength, m_FilePtr);

        
        optionName = strtok(lineBuffer, "=");
        optionValue = strtok(NULL, "=");

        if ( optionName && optionValue ) {
            TrimWhitespace(optionName);

            
            for ( int n = 0; optionValue[n] != '\0'; n++ ) {
                if ( optionValue[n] == '\n' ) {
                    optionValue[n] = '\0';
                    break;
                }
            }

            int optionID = -1;
            for ( int i = 0; i < CONFIG_OPTION_LAST; i++ ) {
                if ( strcasecmp(optionName, option_names[i]) == 0 ) {
                    optionID = i;
                    break;
                }
            }

            if ( optionID != -1 ) {
                assert( optionID >= 0 && optionID < CONFIG_OPTION_LAST);

                
                SetOption(static_cast<CONFIG_OPTION>(optionID), optionValue);
            } else {
                DebugPrint("Found unknown option name: %s\n", optionName);
            }
        }
    }
}


void
ConfigFile::Write()
{
    char lineBuffer[MaxOptionLength];

    
    rewind(m_FilePtr);

    
    OptionMap::const_iterator iter;
    for ( iter = m_Options.begin(); iter != m_Options.end(); iter++ ) {
        snprintf(lineBuffer, MaxOptionLength, "%s=%s\n",
                 option_names[(*iter).first], (*iter).second.c_str());

        DebugPrint("Writing: %s", lineBuffer);

        fputs(lineBuffer, m_FilePtr);
    }

    
    int curPosition = ftell(m_FilePtr);
    ftruncate(fileno(m_FilePtr), curPosition);
}


bool
ConfigFile::IsOptionSet(const CONFIG_OPTION option) const
{
    OptionMap::const_iterator iter;
    for ( iter = m_Options.begin(); iter != m_Options.end(); iter++ ) {
        if ( (*iter).first == option ) {
            return true;
        }
    }

    return false;
}



const char*
ConfigFile::GetOption(const CONFIG_OPTION option) const
{
    OptionMap::const_iterator iter;
    for ( iter = m_Options.begin(); iter != m_Options.end(); iter++ ) {
        if ( (*iter).first == option ) {
            if ( (*iter).second.length() > 0 ) {
                DebugPrint("GetOption: Found value for %s\n", 
                           option_names[option]);
                return (*iter).second.c_str();

            } else {
                
                return NULL;
            }
        }
    }

    return NULL;
}


void
ConfigFile::SetOption(const CONFIG_OPTION option, const char* value)
{
    assert( value );

    m_Options[option] = value;
    DebugPrint("Mapping '%s' to '%s'\n", 
               option_names[option], m_Options[option].c_str());
}


void
ConfigFile::DeleteOption(const CONFIG_OPTION option)
{
    assert( option );

    OptionMap::iterator iter;
    for ( iter = m_Options.begin(); iter != m_Options.end(); iter++ ) {
        if ( (*iter).first == option ) {
            m_Options.erase(iter);
            break;
        }
    }
}


bool
ConfigFile::IsPrintable(char ch)
{
    bool printable;

    switch( ch ) {
    case ' ':
    case '\t':
    case '\n':
        printable = false;
        break;

    default:
        printable = true;
    }

    return printable;
}



void
ConfigFile::TrimWhitespace(char* buffer)
{
    assert( buffer );

    char* firstChar;
    char* lastChar;

    lastChar = buffer + strlen(buffer) - 1;
    while ( !IsPrintable(*lastChar) ) {
        lastChar--;
    }

    firstChar = buffer;
    while ( (firstChar < lastChar) && !IsPrintable(*firstChar) ) {
        firstChar++;
    }

    int strLength = lastChar - firstChar + 1;
    memmove(buffer, firstChar, strLength);
    buffer[strLength] = '\0';
}



void
ConfigFile::DeleteAllOptions()
{
    
    m_Options.clear();
}

