/*
    Copyright (C) SEIKO EPSON CORPORATION 2002-2003. All Rights Reserved.

    3-3-5, Owa, Suwa-shi Nagano-ken, 392-8502 Japan
    Phone: +81-266-52-3131

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


#include <cstring>
#include <cassert>
#include <cerrno>
#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include "debug.h"
#include "netutil.h"
#include "domain.h"

Domain::Domain()
{
	memset( &m_addr, 0, sizeof(m_addr) );
	m_errno = 0;
	m_timeout = -1;
    m_fd = -1;
}

Domain::~Domain()
{
	if ( m_fd >= 0 ) {
		close( m_fd );
	}
}

bool
Domain::Create(SocketType type)
{
    assert( m_fd == -1 );

    if ( type == STREAM ) {
        m_fd = socket(PF_UNIX, SOCK_STREAM, 0);
    } else if ( type == DATAGRAM ) {
        m_fd = socket(PF_UNIX, SOCK_DGRAM, 0);
    } else {
        DebugPrint("Domain: Create called with unknown socket type.\n");
        return false;
    }

	if ( m_fd == -1 ) {
		m_errno = errno;
		return false;
	}

	return true;
}

bool
Domain::Bind( const char path[UNIX_PATH_MAX] )
{
	assert( m_fd >= 0 );
	
	m_addr.sun_family = AF_LOCAL;
	memcpy(m_addr.sun_path, path, UNIX_PATH_MAX);

	if ( bind(m_fd, (struct sockaddr*)&m_addr, sizeof(m_addr)) ) {
		m_errno = errno;
		return false;
	}

	return true;
}

bool
Domain::Connect( const char path[UNIX_PATH_MAX] )
{
	assert( m_fd >= 0 );

	struct sockaddr_un destaddr;
	destaddr.sun_family = AF_LOCAL;
	memcpy(destaddr.sun_path, path, UNIX_PATH_MAX);

    int ret;
    ret = connect(m_fd, (struct sockaddr*)&destaddr, sizeof(destaddr));
    if ( ret == 0 ) {        
        
        DebugPrint("Domain: Connect succeeded\n");
        return true;

    } else {
        assert( ret == -1 );

        if ( m_isNoBlock && errno == EINPROGRESS ) {
            
            struct timeval tv;
            if ( m_timeout >= 0 ) {
                NetworkUtil::TimeToTimeval( m_timeout, &tv );
            }

            fd_set fds;
            FD_ZERO( &fds );
            FD_SET( m_fd, &fds );

            for (;;) {
                if ( m_timeout >= 0 ) {
                    
                    ret = select( m_fd+1, NULL, &fds, NULL, &tv );
                } else {
                    
                    ret = select( m_fd+1, NULL, &fds, NULL, NULL );
                }

                if ( ret == 0 ) {
                    
                    DebugPrint("Domain: Connect attempt timed-out\n");
                    return false;

                } else if ( ret == -1 ) {
                    
                    m_errno = errno;
                    return false;
                }
		
                if ( FD_ISSET(m_fd, &fds) ) {
                    int socketError = 1;
                    socklen_t paramLen = sizeof(socketError);

                    
                    getsockopt(m_fd, SOL_SOCKET, SO_ERROR,
                               &socketError, &paramLen);

                    if ( socketError == 0 ) {
                        
                        DebugPrint("Domain: Nonblocking connect succeeded\n");
                        return true;
                    } else {
                        
                        DebugPrint("Domain: Nonblocking connect failed\n");
                        return false;
                    }
                }
            }

        } else {
            
            m_errno = errno;
            return false;
        }
    }
}

int
Domain::Send( const char* buffer, size_t len )
{
	assert( m_fd >= 0 );

	int ret;
	struct timeval tv;
    if ( m_timeout >= 0 ) {
        NetworkUtil::TimeToTimeval( m_timeout, &tv );
    }

    fd_set fds;
    FD_ZERO( &fds );
    FD_SET( m_fd, &fds );

	for (;;) {
		if ( m_timeout >= 0 ) {
			
            NetworkUtil::TimeToTimeval( m_timeout, &tv );
			ret = select( m_fd+1, NULL, &fds, NULL, &tv );
		} else {
			
			ret = select( m_fd+1, NULL, &fds, NULL, NULL );
		}

		if ( ret == 0 ) {
			
			return 0;

		} else if ( ret == -1 ) {
            
            m_errno = errno;
            return ret;
		}
		
		if ( FD_ISSET(m_fd, &fds) ) {
			
			break;
		}
	}

    int count = 0;
    while (len > 0) {
        ret = send( m_fd, buffer, len, 0 );
        if ( ret == -1 && errno != EAGAIN ) {
            m_errno = errno;
            return count;
        }

        buffer += ret;
        count += ret;
        len -= ret;
    }

	return count;
}

int
Domain::Recv( char* buffer, size_t len )
{
	assert( m_fd >= 0 );

	int ret;

	struct timeval tv;
	if ( m_timeout >= 0 ) {
		NetworkUtil::TimeToTimeval( m_timeout, &tv );
	}

	fd_set fds;
	FD_ZERO( &fds );
	FD_SET( m_fd, &fds );

	for (;;) {
		if ( m_timeout >= 0 ) {
			
			ret = select( m_fd+1, &fds, NULL, NULL, &tv );
		} else {
			
			ret = select( m_fd+1, &fds, NULL, NULL, NULL );
		}

		if ( ret == 0 ) {
			
			return 0;

		} else if ( ret == -1 ) {
			
			m_errno = errno;
			return ret;
		}
		
		if ( FD_ISSET(m_fd, &fds) ) {
			
			break;
		}
	}

    int count = 0;
    while (len > 0) {
        ret = recv( m_fd, buffer, len, 0 );
        if ( ret == -1 && errno != EAGAIN ) {
            m_errno = errno;
            return count;
        }

        buffer += ret;
        count += ret;
        len -= ret;
    }

	return count;
}


void
Domain::SetTimeout( int timeout )
{
	m_timeout = timeout;
}



void
Domain::SetNonBlocking( const bool async )
{
    int flags;

    flags = fcntl(m_fd, F_GETFL, 0);
    if (flags == -1) {
        DebugPrint("Domain: Could not get FD flags\n");
        return;
    }

    if ( async ) {
        flags |= O_NONBLOCK;
    } else {
        flags &= ~O_NONBLOCK;
    }

    if (fcntl(m_fd, F_SETFL, flags) < 0) {
        DebugPrint("Domain: Could not set FD flags\n");
        m_errno = errno;
        return;
    }

	m_isNoBlock = async;
    DebugPrint("Domain: Socket set to non-blocking\n");
}


int
Domain::GetLastError()
{
	int ret = m_errno;
	m_errno = 0;

	return ret;
}
