/*
    Copyright (C) SEIKO EPSON CORPORATION 2002-2003. All Rights Reserved.

    3-3-5, Owa, Suwa-shi Nagano-ken, 392-8502 Japan
    Phone: +81-266-52-3131

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


#include <cassert>
#include <cerrno>
#include <sys/socket.h>
#include <netinet/ip.h>
#include "netutil.h"
#include "icmp.h"

ICMP::ICMP(int fd) : IP(fd)
{
}

ICMP::~ICMP()
{
}

bool 
ICMP::Create()
{
	assert( m_fd == -1 );

	m_fd = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP);
	if ( m_fd == -1 ) {
		m_errno = errno;
		return false;
	}

	return true;
}

int
ICMP::SendTo( const char* buffer, size_t len,
			  in_addr_t addr, unsigned short port )
{
	assert( m_fd >= 0 );

	int ret;

	struct timeval tv;
	if ( m_timeout >= 0 ) {
		NetworkUtil::TimeToTimeval( m_timeout, &tv );
	}

	fd_set fds;
	FD_ZERO( &fds );
	FD_SET( m_fd, &fds );

	for (;;) {
		if ( m_timeout >= 0 ) {
			
			ret = select( m_fd+1, NULL, &fds, NULL, &tv );
		} else {
			
			ret = select( m_fd+1, NULL, &fds, NULL, NULL );
		}

		if ( ret == 0 ) {
			
			return 0;

		} else if ( ret == -1 ) {
			
			m_errno = errno;
			return ret;
		}
		
		if ( FD_ISSET(m_fd, &fds) ) {
			
			break;
		}
	}

	struct sockaddr_in destaddr;

	destaddr.sin_family = AF_INET;
	destaddr.sin_port = htons(port);
	destaddr.sin_addr.s_addr = htonl(addr);

	ret = sendto( m_fd, buffer, len, 0, 
				  (struct sockaddr*)&destaddr, 
				  sizeof(destaddr) );

	if ( ret == -1 ) {
		m_errno = errno;
	}

	return ret;
}

int
ICMP::RecvFrom( char* buffer, size_t len,
				in_addr_t* addr, unsigned short* port )
{
	assert( m_fd >= 0 );

	int ret;

	struct timeval tv;
	if ( m_timeout >= 0 ) {
		NetworkUtil::TimeToTimeval( m_timeout, &tv );
	}

	fd_set fds;
	FD_ZERO( &fds );
	FD_SET( m_fd, &fds );

	for (;;) {
		if ( m_timeout >= 0 ) {
			
			ret = select( m_fd+1, &fds, NULL, NULL, &tv );
		} else {
			
			ret = select( m_fd+1, &fds, NULL, NULL, NULL );
		}

		if ( ret == 0 ) {
			
			return 0;

		} else if ( ret == -1 ) {
			
			m_errno = errno;
			return ret;
		}
		
		if ( FD_ISSET(m_fd, &fds) ) {
			
			break;
		}
	}

	struct sockaddr_in destaddr;
	socklen_t size = sizeof(destaddr);

	ret = recvfrom( m_fd, buffer, len, 0, 
						(struct sockaddr*)&destaddr, 
						&size );

	if ( ret == -1 ) {
		m_errno = errno;
		return ret;
	}

	if ( size != sizeof(destaddr) ) {
		return -1; 
	}

	*addr = ntohl(destaddr.sin_addr.s_addr);
	*port = ntohs(destaddr.sin_port);

	return ret;
}

int
ICMP::CalculateChecksum( unsigned char* buffer, int len )
{
	int sum = 0;
	int remaining = len;
	unsigned short* word = (unsigned short *)buffer;
	unsigned short answer = 0;
	
	
	while ( remaining > 1 ) {
		sum += *word;
		word++;
		remaining -= 2;
	}

	
	if ( remaining ) {
		answer = (*(unsigned char *)word) << 8;
		sum += answer;
	}

	
	sum = (sum >> 16) + (sum & 0xffff);
	sum += (sum >> 16);

	
	answer = ~sum;
	
	return answer;
}
