/*
    Copyright (C) SEIKO EPSON CORPORATION 2002-2003. All Rights Reserved.

    3-3-5, Owa, Suwa-shi Nagano-ken, 392-8502 Japan
    Phone: +81-266-52-3131

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/




#include <unistd.h>
#include <cstdio>
#include <cassert>
#include <cstring>
#include <cerrno>
#include <sys/ioctl.h>
#include <net/if.h>
#include <arpa/inet.h>
#include "validate.h"
#include "netutil.h"
#include "debug.h"
#include "ip.h"
#include "tcp.h"
#include "domain.h"

static const int MAC_ADDRESS_STRLEN = 32;
static const int IPADDR_ADDRESS_STRLEN = 32;


bool
NetworkUtil::StringToAddress(const char* ip_address, in_addr_t* addr)
{
    assert( ip_address );
    assert( addr );

	struct in_addr a;
	
	if ( inet_aton(ip_address, &a) == 0 ) {
		*addr = INADDR_NONE;
		return false;
	}

	
	*addr = ntohl(a.s_addr);
	return true;
}


const char*
NetworkUtil::AddressToString(const unsigned char addr[4])
{
    static char buffer[32];
    
    memset(buffer, 0, 32);
	snprintf(buffer, 32, "%u.%u.%u.%u", addr[0], addr[1], addr[2], addr[3]);

    return buffer;
}


in_addr_t
NetworkUtil::MaskToAddress( int mask )
{
    assert(mask >= 0);
    assert(mask <= 32);

	in_addr_t addr = 0;

	while (mask > 0) {
		
		addr = (addr >> 1) | 0x80000000;
		mask--;
	}

	return addr;
}



in_addr_t
NetworkUtil::MakeAddress(const unsigned char buf[4])
{
	const char* tmp;
	in_addr_t addr;

    tmp = AddressToString(buf);
	StringToAddress( tmp, &addr );

	return addr;
}



void
NetworkUtil::PrintSocketError(IP& sock)
{
    DebugPrint( "IP socket error: %s\n", strerror(sock.GetLastError()) );
}



void
NetworkUtil::PrintSocketError(Domain& sock)
{
    DebugPrint( "Domain socket error: %s\n", strerror(sock.GetLastError()) );
}



bool
NetworkUtil::IsSameAddress(const unsigned char* addr1, 
                           const unsigned char* addr2)
{
    assert(addr1);
    assert(addr2);

    return (memcmp(addr1, addr2, 4) == 0);
}



void
NetworkUtil::TimeToTimeval( int ms, struct timeval *tv )
{
	assert( ms >= 0 );
	assert( tv != 0 );

	tv->tv_sec = ms / 1000;
	tv->tv_usec = (ms % 1000) * 1000;
}



const char*
NetworkUtil::GetHardwareAddress(const char* device)
{
    static char deviceMAC[MAC_ADDRESS_STRLEN];

    
    TCP tcpSocket;
    if ( ! tcpSocket.Create() ) {
        return 0;
    }

    
    struct ifreq devReq;
    strncpy(devReq.ifr_name, device, IFNAMSIZ);

    int status = ioctl(tcpSocket.GetFD(), SIOCGIFHWADDR, &devReq);
    if ( status == -1 ) {
        DebugPrint("%s: %s\n", devReq.ifr_name, strerror(errno));
        return 0;
    }

    
    sprintf(deviceMAC, "%2.2X:%2.2X:%2.2X:%2.2X:%2.2X:%2.2X",
            (unsigned char)devReq.ifr_hwaddr.sa_data[0],
            (unsigned char)devReq.ifr_hwaddr.sa_data[1],
            (unsigned char)devReq.ifr_hwaddr.sa_data[2],
            (unsigned char)devReq.ifr_hwaddr.sa_data[3],
            (unsigned char)devReq.ifr_hwaddr.sa_data[4],
            (unsigned char)devReq.ifr_hwaddr.sa_data[5] );

    return deviceMAC;
}

in_addr_t
NetworkUtil::GetInterfaceAddress(const char* device)
{
    sockaddr_in* saddr;

    
    TCP tcpSocket;
    if ( ! tcpSocket.Create() ) {
        return 0;
    }

    
    struct ifreq devReq;
    strncpy(devReq.ifr_name, device, IFNAMSIZ);

    int status = ioctl(tcpSocket.GetFD(), SIOCGIFADDR, &devReq);
    if ( status == -1 ) {
        DebugPrint("%s: %s\n", devReq.ifr_name, strerror(errno));
        return 0;
    }

    saddr = (sockaddr_in*)&devReq.ifr_addr;
    return saddr->sin_addr.s_addr;
}


bool
NetworkUtil::IsValidNetAddress(in_addr_t addr, in_addr_t subnet)
{
    
    if ( subnet == 0 ) {
        return true;
    }

    if ( (addr & subnet) != addr &&
         (addr & subnet) != 0    && 
         (addr & ~subnet) != 0   &&
         ((addr & ~subnet) | subnet) != 0xFFFFFFFF ) 
    {
        return true;
    }

    return false;
}
